'''
This module handles the widgets in the first tab of the GUI, which concern the loading of a image analysis / fcs analysis directory.
It also contains the widgets for bead normalization, he license display button, and GUI theme changes.



This file is licensed under the GPL3 license. No significant portion of the code here is known to be derived from another project 
(in the sense of needing to be separately / simulataneously licensed)
'''
import os
import sys
import shutil
from typing import Union
from pathlib import Path
import tkinter as tk
import zipfile

import requests
import customtkinter as ctk
import numpy as np
# import pandas as pd
import matplotlib.pyplot as plt
from PIL import Image


from .._vendor import fcsparser

from ..Utils.sharedClasses import Project_logger, Analysis_logger, CtkSingletonWindow, overwrite_approval      # , DirectoryDisplay
from .bead_norm import CyTOF_bead_normalize
from ..ImageProcessing.ImageAnalysisClass import direct_to_Analysis, toggle_in_gui, imc_entrypoint
from ..ImageProcessing.ImageAnalysisWidgets import ImageProcessingWidgets 
from ..Analysis_widgets.Spatial_GUI import Spatial_py  
from ..Analysis_widgets.Analysis_GUI import Analysis_py_widgets 
from ..Pixel_Classification.Classifiers_GUI import Pixel_class_widgets
from ..Pixel_Classification.use_classifiers_GUI import Pixel_usage_widgets

__all__ = ["fetch_CyTOF_example", 
           "fetch_IMC_example"]

homedir = __file__.replace("\\","/")
homedir = homedir[:(homedir.rfind("/"))]
## do it twice to get up to the top level directory:
homedir = homedir[:(homedir.rfind("/"))] 
Theme_link = homedir + '/Assets/theme.txt'

chosen_dir_file = homedir + "/Assets/dir_choice.txt"
with open(chosen_dir_file) as file:
    chosen_dir = file.read()
if chosen_dir == 'None"' :
    chosen_dir = homedir[:(homedir.find("/")) + 1]

class App(ctk.CTk):
    '''
    This is the main window for the GUI. It also contains the Tabholder class, which coordinates the tabs of the program
    '''
    def __init__(self, backend_process = None):
        super().__init__()
        self.napari_launch = backend_process
        self.light_dark = "dark"
        toggle = toggle_in_gui()
        if not toggle:    ## this horrible little construct ensures _in_gui is True even if reinitialized
            toggle_in_gui()
        ctk.set_appearance_mode("dark")
        if sys.platform == "win32":
            self.iconbitmap(homedir + "/Assets/Capture.ico")    ## Thanks to: https://www.freeconvert.com/jpg-to-ico for converting .jpg to 
                                                                            # .ico file
        with open(Theme_link) as theme:
            self.theme = theme.read()
        if len(self.theme) == 0:
            self.theme = "blue"   ## this is the default theme for ctk
        elif (self.theme == "green") or (self.theme == "blue"):
            ctk.set_default_color_theme(self.theme)       ## green and blue are themes bundled with customtkinter (don't require a link)
        else:
            theme_dir = homedir +  "/Assets/ctkThemeBuilderThemes/"
            ctk.set_default_color_theme(theme_dir + self.theme + ".json")

        ### The 1200 by 1920 ratio is from the computer I was using to develop this program:
        dev_comp_height = 1200
        dev_comp_width = 1920
        instance_height = self.winfo_screenheight()
        instance_width = self.winfo_screenwidth()
        ratio_height = instance_height / dev_comp_height
        ratio_width = instance_width / dev_comp_width
        self.ratio = min(ratio_height,ratio_width)
        self.scaling = 1.1
        ctk.set_widget_scaling(self.ratio * 1.1)
        ctk.set_window_scaling(self.ratio * 1.1)   ### originally, PALMETOBUG was made into a rather small window, so scaling up is usual

        self.geometry("1600x1000+0+0")
        self.resizable(True,True)
        self.title("PalmettoBUG")

        self.Tabs = self.Tabholder(self)
        self.entrypoint = EntryPoint(self.Tabs)
        self.entrypoint.grid(row = 0, column = 0, padx = 10, pady = 10)

    def re__init__(self) -> None:
        self.Tabs.destroy()
        self.Tabs = self.Tabholder(self)
        self.entrypoint.destroy()
        self.entrypoint = EntryPoint(self.Tabs)
        self.entrypoint.grid(row = 0, column = 0, padx = 10, pady = 10)

    def destroy(self) -> None:   ## overwrite base class destroy method to ensure that matplotlib is closed
                         ## modified directly from the original customtkinter code for ctk.CTk.destroy()
        plt.close()
        self._disable_macos_dark_title_bar()

        # call destroy methods of super classes
        tk.Tk.destroy(self)
        ctk.windows.widgets.appearance_mode.CTkAppearanceModeBaseClass.destroy(self)
        ctk.windows.widgets.scaling.CTkScalingBaseClass.destroy(self)

    class Tabholder(ctk.CTkTabview):
        def __init__(self, master):
            super().__init__(master)
            self.start = self.add('Start')
            self.start.directory = None

            self.mcdprocessing = self.add('MCD / Image Processing')
            self.Pixel_Classification = self.add("Pixel Classification")

            self.py_exploratory = self.add('Analysis')

            self.Spatial = self.add('Spatial Analysis')
            self.set('Start')

            self.configure(width = 1550, height = 950)
            self.grid(row = 0, column = 0)

            self.px_classification = self.Px_secondary_Tabholder(self.Pixel_Classification)
            self.px_classification.grid()

            self.py_exploratory.analysiswidg = Analysis_py_widgets(self.py_exploratory)
            self.py_exploratory.analysiswidg.grid()

            self.image_proc_widg = ImageProcessingWidgets(self.mcdprocessing)
            self.image_proc_widg.grid()

            self.Spatial.widgets = Spatial_py(self.Spatial)
            self.Spatial.widgets.grid()

        class Px_secondary_Tabholder(ctk.CTkTabview):
            def __init__(self, master):
                super().__init__(master)
                self.master = master
                self.create = self.add("Create Pixel Classification")
                self.use_class = self.add("Use Pixel Classification outputs")
                self.set('Create Pixel Classification')
                self.configure(width = 1500, height = 925)
                self.grid(row = 0, column = 0)
                self.create.px_widg = Pixel_class_widgets(master = self.create)  
                self.create.px_widg.grid()
                self.use_class.px_widg = Pixel_usage_widgets(master = self.use_class) 
                self.use_class.px_widg.grid()

class EntryPoint(ctk.CTkFrame):
    '''
    This is the initial frame that helps the user provide the directory for data ingestion.

    It can always be returned to by the user to re-start the data ingestion / processing. 
    '''
    def __init__(self, Tabholder: App.Tabholder):
        super().__init__(Tabholder.start)
        self.master = Tabholder   ##### so: self.master.master  == the main App
        self.directory = None
        self.image_proc_widg = self.master.image_proc_widg
        self.configure(height = 400, width = 1600)

        ## widget set for the MCD entry point
        label = ctk.CTkLabel(master = self, text = "Directory structure options:" 
                                    "\n\n 1). A Directory containing a /raw folder with .mcd files inside it"
                                    "\n  or .tiff / .ome.tiff files inside it"
                                    "\n\n 2). a Directory containing a /main/Analysis_fcs folder \n with  .fcs files inside it")

        label.grid(column = 0, row = 0, padx = 2, pady = 2, rowspan = 3)
        label.configure(anchor = "w")

        #button_MCD = ctk.CTkButton(master = self, 
        #                text = "Choose MCD directory and begin", 
        #                command = lambda: self.img_entry_func("", from_mcds = True))
        #button_MCD.grid(column = 1, row = 1, padx = 10, pady = 10)

        ## widget set for the images-only entry point
        button_img = ctk.CTkButton(master = self, 
                    text = "Choose Image directory and begin", 
                    command = lambda: self.img_entry_func("", from_mcds = None))
        button_img.grid(column = 1, row = 1, padx = 10, pady = 3)

        ## widget set for FCS entry point
        button_fcs = ctk.CTkButton(master = self, text = "Choose FCS directory and begin", command = lambda: self.FCS_choice(""))
        button_fcs.grid(column = 1, row = 2, padx = 10, pady = 3)

        button_example = ctk.CTkButton(master = self, text = "Load Example Data", command = self.launchExampleDataWindow)
        button_example.grid(column = 0, row = 4, padx = 10, pady = 10) 

        ## The widget for the entry of the resolution of the images (can ignore for CyTOF / solution mode analyses)
        self.X_Y_entry = self.X_Y_res_frame(self)
        self.X_Y_entry.grid(column = 1, row = 4, padx = 10, pady = 10)

        '''
        label_dir_disp = ctk.CTkLabel(master = self, text = "Alternatively, select the folder \n using the directory display below:")
        label_dir_disp.grid(column = 0, row = 5, padx = 10, pady = 10)

        self.dir_disp = DirectoryDisplay(self)
        # self.dir_disp.grid(column = 0, row = 6, padx = 3, pady = 3, rowspan = 4)
        # try:
        #    self.dir_disp.setup_with_dir(chosen_dir) 
        # except FileNotFoundError:
        #    root_dir = homedir[:(homedir.find("/")) + 1]
        #    self.dir_disp.setup_with_dir(root_dir)

        self.dir_helper = directory_display_helper_frame(self, self.dir_disp)
        # self.dir_helper.grid(column = 0, row = 10, padx = 3, pady = 10, rowspan = 3)
        '''

        label = ctk.CTkLabel(master = self, 
                             text = "Good Evening! I am the distinguished Mr. Palmett O. Bug, Esq." 
                                    "and gentleman. \n I am pleased to make your acquaintance. ")
        label.grid(column = 2, row = 0, padx = 5, pady = 5)

        self.display = self.display_image_button(self)
        self.display.grid(column = 2, row = 1, padx = 5, pady = 5, rowspan = 18)

        normalizer_button = ctk.CTkButton(master = self,text = "Launch Normalizer", command = lambda: self.normalize_fcs_choice(""))
        normalizer_button.grid(column = 2, row = 20, padx = 5, pady = 5)

        buttonConfig = ctk.CTkButton(master = self, text = "Launch GUI configuration window", command = self.call_configGUI)
        buttonConfig.grid(column = 3, row = 1, padx = 10, pady = 10)
        buttonConfig.configure(anchor = "e")

        def show_GPL() -> None:
            GPL_window(self)

        button_baby = ctk.CTkButton(master = self, text = "See LICENSE Details", command = show_GPL)
        button_baby.grid(column = 3, row = 2, padx = 10, pady = 10) 
    
        self.after(200, lambda: self.focus())

    class display_image_button(ctk.CTkButton):
        def __init__(self, master):
            super().__init__(master)
            self.master = master
            image = Image.open(homedir + "/Assets/Capture3.jpg")
            self.configure(text = "", image = ctk.CTkImage(image, 
                                            size = (500,500)), 
                                            height = 550, 
                                            width = 550, 
                                            fg_color = "white", 
                                            hover = "white")

        def update_image(self, image: Union[str, ctk.CTkImage]) -> None:
            if isinstance(image, str):
                image = Image.open(image)
                image = ctk.CTkImage(image, size = (550,550))
            self.configure(image = image)

    def call_configGUI(self) -> None:
        configGUI_window(self.master.master)

    def launchExampleDataWindow(self):
        LoadExampleDataWindow(self)

    class X_Y_res_frame(ctk.CTkFrame):
        '''
        The widget frame containing the entries for the X & Y resolutions of the images (can ignore if using CyTOF / solution mode)
        '''
        def __init__(self, master):
            super().__init__(master)
            self.master = master
            label = ctk.CTkLabel(master = self, text = "Enter X and Y resolution (in micrometers):")
            label.grid(column = 0, row = 0, columnspan = 2, padx = 10, pady = 10)
            
            self.entry_X = ctk.CTkEntry(master = self, textvariable = ctk.StringVar(value = "1.0"))
            self.entry_X.grid(column = 0, row = 1, padx = 10, pady = 10)

            self.entry_Y = ctk.CTkEntry(master = self, textvariable = ctk.StringVar(value = "1.0"))
            self.entry_Y.grid(column = 1, row = 1, padx = 10, pady = 10)

    def img_entry_func(self, directory: Union[str, None] = None, resolutions = None, from_mcds = None) -> None:
        '''
        This function directs the directory (and resolutions) into an experiment -- with MCD initial read-in -- 
        and sets up the Image processing Widgets 
        '''
        if resolutions is None:
            try:
                resX = float(self.X_Y_entry.entry_X.get())
                resY = float(self.X_Y_entry.entry_Y.get())
            except ValueError:
                tk.messagebox.showwarning("Warning!", message = "Resolution X / Y must be numbers!")
                return
        else:
            resX = float(resolutions[0])
            resY = float(resolutions[1])

        self.master.directory = directory.replace("\\","/")
        ### don't want the current directory to be searched (when entry field is blank)....
        if self.master.directory is None:
            self.master.directory = "not a directory"   ## if the directory remains None, odd behaviours can result 
                                                            # because of default arguments in downstram functions ---> 
                                                            # like the current directory being searched
        if len(self.master.directory) == 0:
            self.master.directory = tk.filedialog.askdirectory()
            if self.master.directory == "":
                return

        ## This is a check of the entered directory existing:
        try:
            files = [i for i in os.listdir(self.master.directory + "/raw") if ((i.lower().find(".mcd") != -1) or (i.lower().find(".tif") != -1))]
            if len(files) == 0:
                tk.messagebox.showwarning("Warning!", message = "There are no files in the raw folder!")
                return
        except FileNotFoundError:
            tk.messagebox.showwarning("Warning!", message = "This is not a valid directory!")
            return
        
        if from_mcds is None:
            example_files = sorted(os.listdir(self.master.directory + "/raw"))
            extensions = [i[(i.rfind(".") + 1):].lower() for i in example_files]
            if ("mcd" in extensions) and ("tif" not in extensions) and ("tiff" not in extensions):
                from_mcds = True
            elif ("tiff" in extensions) or ("tif" in extensions):
                from_mcds = False
            else:
                tk.messagebox.showwarning("Warning!", message = "The /raw sub-folder must contain .mcd or .tiff files, and ONLY .mcd or ONLY .tiff files (not a mixture)!")
                return

        Experiment = imc_entrypoint(directory = self.master.directory, 
                                      resolutions = [resX, resY], 
                                      from_mcds = from_mcds)

        self.master.px_classification.create.px_widg.set_directory(Experiment.directory_object)
        self.master.px_classification.use_class.px_widg.set_directory(Experiment.directory_object)
        self.master.px_classification.use_class.px_widg.add_experiment(experiment = Experiment)

        ## this removes any old widgets of a previously entered  MCD directory:
        if self.image_proc_widg is not None:
            self.image_proc_widg.destroy()
            self.image_proc_widg = ImageProcessingWidgets(self.master.mcdprocessing)
            self.image_proc_widg.grid(column = 0, row = 0)
        else:
            self.image_proc_widg = ImageProcessingWidgets(self.master.mcdprocessing)
            self.image_proc_widg.grid(column = 0, row = 0)

        self.image_proc_widg.add_Experiment(Experiment, from_mcds = from_mcds)

        # Initialize the image processing widgets:                 
        self.image_proc_widg.initialize_buttons(self.master.directory) 

        ## set up project logger once setup is successful:
        project_log = Project_logger(self.master.directory).return_log()
        project_log.info(f"Start log in directory {self.master.directory}/Logs after loading from MCD files")
        self.master.set('MCD / Image Processing')

    def normalize_fcs_choice(self, directory: Union[None, str] = None) -> None:
        self.master.py_exploratory.X = self.X_Y_entry.entry_X.get()
        self.master.py_exploratory.Y = self.X_Y_entry.entry_Y.get()
        self.master.directory = directory.replace("\\","/")

        if directory == "":
            self.master.directory = tk.filedialog.askdirectory()
            if self.master.directory == "":
                return
        try:
            beads_files = [i for i in os.listdir(self.master.directory + "/beads") if i.lower().find(".fcs") != -1]
            non_beads_files = [i for i in os.listdir(self.master.directory + "/no_beads") if i.lower().find(".fcs") != -1]
            if len(beads_files) == 0:
                tk.messagebox.showwarning("Warning!", message = "There are no FCS files in the /no_beads folder!")
                return
            if len(non_beads_files) == 0:
                tk.messagebox.showwarning("Warning!", message = "There are no FCS files in the /beads folder!")
                return
        except FileNotFoundError:
            tk.messagebox.showwarning("Warning!", message = "This is not a valid directory!")
            return 
        
        ### Launch a window for taking in user inputs as to which channels are beads channels and which are channels to normalize

        bead_1 = self.master.directory + "/beads/" + beads_files[0]
        _, beads_1 = fcsparser.parse(bead_1, channel_naming = "$PnS")
        channels = beads_1.columns
        # print(channels)
        self.channels_to_use = Channel_normalization_window(self, channels, self.master.directory)

    def FCS_choice(self, directory: Union[None, str] = None) -> None:
        self.master.py_exploratory.X = self.X_Y_entry.entry_X.get()
        self.master.py_exploratory.Y = self.X_Y_entry.entry_Y.get()

        self.master.directory = directory.replace("\\","/")

        if self.master.directory is None:
            self.master.directory = "not a directory"
        if len(self.master.directory) == 0:
            self.master.directory = tk.filedialog.askdirectory()
            if self.master.directory == "":
                return
        try:
            fcs_files = [i for i in os.listdir(self.master.directory + "/main/Analysis_fcs") if i.lower().find(".fcs") != -1]
            if len(fcs_files) == 0:
                tk.messagebox.showwarning("Warning!", message = "There are no files in the /main/Analysis_fcs folder!")
                return
        except FileNotFoundError:
            tk.messagebox.showwarning("Warning!", message = "This is not a valid directory!")
            return   
        
        ## set up project logger:
        project_log = Analysis_logger(self.master.directory + "/main").return_log()
        direct_to_Analysis(self.master, self.master.directory)
        project_log.info(f"Start log of experiment from the directory {self.master.directory + '/main'}/Logs after loading .fcs for direct analysis")
        self.master.set('Analysis')
        self.master.master.Tabs.py_exploratory.analysiswidg.setup_dir_disp(self.master.directory + "/main")  
        self.master.master.Tabs.Spatial.widgets.setup_dir_disp(self.master.directory + "/main")


class configGUI_window(ctk.CTkToplevel, metaclass = CtkSingletonWindow):

    def __init__(self, App_instance: App):
        super().__init__()
        self.master = App_instance
        self.lt_drk = App_instance.light_dark

        label = ctk.CTkLabel(master = self, text = "Make changes to the GUI appearance:")
        label.grid(padx = 5, pady = 5)

        label1 = ctk.CTkLabel(master = self, text = "Select a Ratio to change the sizes of the widgets & window:")
        label1.grid(padx = 5, pady = 5)

        self.slider = ctk.CTkComboBox(master = self, 
                        values = ["0.85","0.9","0.95","1.0","1.05","1.1","1.15","1.2","1.25","1.3"], 
                        command = self.slider_moved)
        self.slider.grid(padx = 5, pady = 5)
        self.slider.set(App_instance.scaling)

        self.theme_dir = homedir +  "/Assets/ctkThemeBuilderThemes"
        to_display = [str(i).replace(".json","").replace("\\","/") for i in Path(self.theme_dir).rglob("[!.]*.json")]
        to_display = ["green","blue"] + [i[(i.rfind("/") + 1):] for i in to_display] 

        label2 = ctk.CTkLabel(master = self, text = "Change the color theme (note this may reset unsaved progress in an analysis):")
        label2.grid(padx = 5, pady = 5)

        self.combobox = ctk.CTkComboBox(master = self, values = to_display, command = self.change_theme)
        self.combobox.grid(padx = 5, pady = 5)
        self.combobox.set(App_instance.theme)

        self.light_dark = ctk.CTkButton(master = self, text = "Toggle theme light / dark", command = self.toggle_light_dark)
        self.light_dark.grid(padx = 5, pady = 5)

        self.after(200, lambda: self.focus())

    def re__init__(self) -> None:
        Appinstance = self.master
        self.destroy()
        configGUI_window(Appinstance)
    
    def toggle_light_dark(self) -> None:
        if self.master.light_dark == "dark":
            ctk.set_appearance_mode("light")
            self.master.light_dark = "light"
        elif self.master.light_dark == "light":
            ctk.set_appearance_mode("dark")
            self.master.light_dark = "dark"
        self.after(200, lambda: self.focus())

    def slider_moved(self, scaling: float) -> None:
        scaling = float(scaling)
        ctk.set_widget_scaling(self.master.ratio * scaling)
        self.master.scaling = scaling

    def change_theme(self, new_theme: str) -> None:
        if new_theme in ["green", "blue"]:
            ctk.set_default_color_theme(new_theme) 
        else:
            ctk.set_default_color_theme(self.theme_dir + f"/{new_theme}.json") 
        with open(Theme_link, mode = 'w') as theme:
            theme.write(new_theme)
        self.master.theme = new_theme
        self.master.re__init__()
        self.re__init__()

class GPL_window(ctk.CTkToplevel, metaclass = CtkSingletonWindow):

    def __init__(self, master):
        super().__init__(master)
        self.alt_license = None
        license_dir = homedir + "/Assets/LICENSE.txt"
        with open(license_dir) as file:
            self.license = file.read()

        label = ctk.CTkLabel(master = self, text = """Copyright Medical University of South Carolina 2024-2025 \n
                             
        While this project as a whole and all its scripts are licensed here under GPL-3 (see text box below)
        Much of this project is also heavily based on / derived from a number of existing open-source packages:
            \t Steinbock (https://github.com/BodenmillerGroup/steinbock), CATALYST (https://github.com/HelenaLC/CATALYST)
            \t Diffcyt (https://github.com/lmweber/diffcyt), QuPath (https://github.com/qupath/qupath/)
            \t opencv (https://github.com/opencv/opencv), spatstat (https://github.com/spatstat/spatstat.core), 
            \t spaceanova (https://github.com/sealx017/SpaceANOVA/), Premessa / Bead-normalization  (https://github.com/ParkerICI/premessa / https://github.com/nolanlab/bead-normalization)
            \t singletons (https://github.com/jmaroeder/python-singletons ), scikit-image (https://github.com/scikit-image/scikit-image),
            \t ctk_theme_builder (https://github.com/avalon60/ctk_theme_builder)

        Additionally, some packages' code is directly copied, with some modification inside PalmettoBUG itself ("vendored"):
            Vendored packages == fcsparser, fcsy, pyometiff, qnorm, readimc, sigfig, and steinbock  (see _vendor folder inside PalmettoBUG package)

        The listed packages are noted because of their use beyond merely importation / use of API / use of documentation files for the many library dependencies of this project,
        and because of their extensive use in different parts of the program. See individual .py files for details & any other packages that parts of the program may have derived from.
        As far as I am aware, all imported libraries at runtime are themselves permissively or GPL-compatibly licensed, with the possible exception of opencv on MacOS and 
        Linux systems only -- opencv-python on those OS is bundled with OpenSSL 1.1.1, which is free & open-source but GPL incompatible. However, the OpenSSL library should not be 
        used / imported by PalmettoBUG at runtime.
            
        NOTE! The isoSegDenoise sister program is not licensed under GPL and contain segmentation options with non-commercial / academic
        use only restrictions. It is launched from PalmettoBUG in a separate process through command line. 
            """, 
            anchor = 'w', 
            justify = 'left')
        
        label.grid(pady = 5, padx = 5, column = 0, row = 0, columnspan = 2)

        self.button_main_license_text = ctk.CTkButton(master = self, 
                                                text = "Display Main License (GPL-3)", 
                                                state = 'disabled', 
                                                command = self.display_main)
        self.button_main_license_text.grid(pady = 5, padx = 5, column = 0, row = 1)

        self.button_alt_license_text = ctk.CTkButton(master = self, 
                                                text = "Display 3rd Party license details", 
                                                command = self.display_3rd)
        self.button_alt_license_text.grid(pady = 5, padx = 5, column = 1, row = 1)

        self.display = ctk.CTkTextbox(master = self, activate_scrollbars = True, wrap = 'none', width = 700, height = 500)
        self.display.grid(column = 0, row = 2)

        self.display.insert("0.0", self.license)
        self.display.configure(state = "disabled")

        self.after(200, lambda: self.focus())

    def display_main(self):
        ''''''
        self.button_alt_license_text.configure(state = 'normal')
        self.display.configure(state = 'normal')
        self.display.delete("0.0","end")
        self.display.insert("0.0", self.license)
        self.display.configure(state = "disabled")
        self.button_main_license_text.configure(state = 'disabled')

    def display_3rd(self):
        ''''''
        if self.alt_license is None:
            license_dir = homedir + "/Assets/Other_License_Details.txt"
            with open(license_dir) as file:
                self.alt_license = file.read()
        self.button_main_license_text.configure(state = 'normal')
        self.display.configure(state = 'normal')
        self.display.delete("0.0","end")
        self.display.insert("0.0", self.alt_license)
        self.display.configure(state = "disabled")
        self.button_alt_license_text.configure(state = 'disabled')

class Channel_normalization_window(ctk.CTkToplevel, metaclass = CtkSingletonWindow):
    def __init__(self, master, channels: list[str], directory: str):
        super().__init__(master)
        self.master = master
        self.directory = directory
        self.channels = channels

        self.scrollable_frame = ctk.CTkScrollableFrame(master = self, width = 350, height = 500)
        self.scrollable_frame.grid()

        self.bead_label = ctk.CTkLabel(master = self.scrollable_frame, text = str("Select \n Bead Channels:"))
        self.bead_label.grid(row = 1, column = 1, pady = 3, padx = 3)

        self.to_norm_label = ctk.CTkLabel(master = self.scrollable_frame, text = str("Select Channels \n to normalize:"))
        self.to_norm_label.grid(row = 1, column = 2, pady = 3, padx = 3)

        self.checkbox_beads_list = []
        self.checkbox_norm_list = []
        for ii,i in enumerate(channels):
            label = ctk.CTkLabel(master = self.scrollable_frame, text = str(i))
            label.grid(row = ii + 2, column = 0, pady = 3, padx = 3)

            checkbox_beads = ctk.CTkCheckBox(master = self.scrollable_frame, text = "", onvalue = True, offvalue = False)
            checkbox_beads.grid(row = ii + 2, column = 1, pady = 3, padx = 3)
            self.checkbox_beads_list.append(checkbox_beads)

            checkbox_to_norm = ctk.CTkCheckBox(master = self.scrollable_frame, text = "", onvalue = True, offvalue = False)
            checkbox_to_norm.grid(row = ii + 2, column = 2, pady = 3, padx = 3)
            if str(i) != "Time":
                checkbox_to_norm.select()
            self.checkbox_norm_list.append(checkbox_to_norm)

        label_straight = ctk.CTkLabel(master = self, 
            text = "Check to go straight to an analysis after normalization"
                "\nYou will be prompted first to select an empty directory \n to set up the new analysis in")
        label_straight.grid(column = 3)

        self.straight_to_analysis = ctk.CTkCheckBox(master = self, text = "", onvalue = True, offvalue = False)
        self.straight_to_analysis.grid(column = 3)

        button = ctk.CTkButton(master = self, 
                            text = "Perform Normalization!", 
                            command = lambda: self.accept_and_normalize(channels = self.retrieve_channels(),
                                                                    to_analysis = self.straight_to_analysis.get()))
        button.grid(row = ii + 3, column = 2, pady = 3, padx = 3)
        self.after(200, self.focus())

    def retrieve_channels(self) -> tuple[list, list]:
        bead_channels = np.array([i.get() for i in self.checkbox_beads_list])
        to_norm_channels = np.array([i.get() for i in self.checkbox_norm_list])
        bead_channels = self.channels[bead_channels]
        to_norm_channels = self.channels[to_norm_channels]
        return bead_channels, to_norm_channels

    def accept_and_normalize(self, channels: list[list[str], list[str]], to_analysis: bool = False) -> None:
        bead_channels, channels_to_normalize = channels
        if not overwrite_approval(self.directory + "/normalization", file_or_folder = "folder"):
            return

        if to_analysis:
            output_directory = self.norm_to_Analysis1()
            if output_directory == "":
                return
            if not overwrite_approval(output_directory + "/main/Analysis_fcs", file_or_folder = "folder"):
                return
        #### move functions into a separate file, likely, as done previously:
        CyTOF_bead_normalize(self.directory + "/beads", 
                                  self.directory + "/no_beads", 
                                  self.directory + "/normalization", 
                                  bead_channels, 
                                  channels_to_normalize = channels_to_normalize, 
                                  include_figures = True)
        if to_analysis:
          self.norm_to_Analysis2(self.directory + "/normalization", output_directory)
    
        self.destroy()

    def norm_to_Analysis1(self):
        output_dir = tk.filedialog.askdirectory()
        return output_dir
    
    def norm_to_Analysis2(self, norm_directory, output_directory):
        output_analysis_fcs = output_directory + "/main"
        if not os.path.exists(output_analysis_fcs):
            os.mkdir(output_analysis_fcs)
        shutil.copytree(norm_directory + "/normalized", output_analysis_fcs + "/Analysis_fcs")
        self.master.FCS_choice(output_directory)

class LoadExampleDataWindow(ctk.CTkToplevel, metaclass = CtkSingletonWindow):
    '''
    '''
    def __init__(self, master):
        super().__init__(master)
        self.master = master
        self.dir = None

        grandlabel = ctk.CTkLabel(master = self, text = "Load Example data into a chosen directory:\n (chosen directory should usually be empty)")
        grandlabel.grid(padx = 3, pady = 3, column = 0, row = 0, columnspan = 2)

        label = ctk.CTkLabel(master = self, text = "Choose Directory:")
        label.grid(padx = 3, pady = 3, column = 0, row = 1)

        self.entry = ctk.CTkEntry(master = self, textvariable = ctk.StringVar(value = ""))
        self.entry.grid(padx = 3, pady = 3, column = 1, row = 1)

        button = ctk.CTkButton(master = self, text = "Launch directory selection", command = self.launch_dir_search)
        button.grid(padx = 3, pady = 3, column = 0, row = 2, columnspan = 2)

        spacer = ctk.CTkLabel(master = self, text = "")
        spacer.grid(padx = 3, pady = 3, column = 0, row = 3)

        button_CyTOF = ctk.CTkButton(master = self, text = "Load CyTOF (solution-mode) data", command = self.load_CyTOF)
        button_CyTOF.grid(padx = 3, pady = 3, column = 0, row = 4)

        button_IMC = ctk.CTkButton(master = self, text = "Load IMC (image-mode) data", command = self.load_IMC)
        button_IMC.grid(padx = 3, pady = 3, column = 1, row = 4)

        self.after(750, self.focus())

    def launch_dir_search(self):
        self.dir = tk.filedialog.askdirectory()
        self.entry.configure(textvariable = ctk.StringVar(value = self.dir))
        self.after(200, self.focus())

    def load_CyTOF(self):
        choice = self.entry.get()
        if choice == "":
            tk.messagebox.showwarning("No Directory Entered!", message = "No Directory Entered, please select one first!")
            return
        if not overwrite_approval(choice + "/main/Analysis_fcs", file_or_folder = "folder", custom_message = "Are you sure you want to overwrite files in this folder"
                                                                                                             "as well as the associated panel and metdata files?"):
            return
        fetch_CyTOF_example(choice)
        self.master.FCS_choice(choice + "/Example_CyTOF")
        self.withdraw()

    def load_IMC(self):
        choice = self.entry.get()
        if choice == "":
            tk.messagebox.showwarning("No Directory Entered!", message = "No Directory Entered, please select one first!")
            return
        if not overwrite_approval(choice + "/raw", file_or_folder = "folder", custom_message = "Are you sure you want to overwrite files in this folder \n "
                                                                                                "as well as the associated panels and metdata files?"):
            return
        fetch_IMC_example(choice)
        self.master.img_entry_func(choice + "/Example_IMC", from_mcds = None)
        self.withdraw()

def fetch_CyTOF_example(new_directory):
    '''
    This copies the CyTOF example data to the supplied new_directory, so that it is ready to loaded. As in: ::

        new_directory = ".../.../..."
        palmettobug.fetch_CyTOF_example(new_directory = new_directory)
        Analysis = palmettobug.Analysis()
        Analysis.load_data(new_directory)

    Calling this fetch function on an existing directory will cause the existing directory's /Analysis_fcs folder to be emptied and refilled with the example data!   
    '''
    if not os.path.exists(new_directory):
        raise Exception("Target directory does not exist")
    CyTOF_data = requests.get("https://zenodo.org/records/14983582/files/Example_CyTOF.zip?download=1")
    with open(new_directory + "/CyTOF_data.zip", 'wb') as write_to:
        write_to.write(CyTOF_data.content)
    zip_archive = zipfile.ZipFile(new_directory + "/CyTOF_data.zip")
    zip_archive.extractall(new_directory)

def fetch_IMC_example(new_directory):
    '''
    This copies the IMC example data to the supplied new_directory, so that it is ready to loaded by, As in: ::

        new_directory = ".../.../..."
        palmettobug.fetch_IMC_example(new_directory = new_directory)
        ImageAnalysis = palmettobug.ImageAnalysis(new_directory, [1.0,1.0], from_mcs = False)

    Calling this fetch function on an existing directory will caues the files in the /raw sub-folder ot be replaced with the exmaple data!
    '''
    if not os.path.exists(new_directory):
        raise Exception("Target directory does not exist")
    IMC_data = requests.get("https://zenodo.org/records/14983582/files/Example_IMC.zip?download=1", stream = True)
    with open(new_directory + "/IMC_data.zip", 'wb') as write_to:
        for i in IMC_data.iter_content(chunk_size = 1024**2):
            write_to.write(i)
    zip_archive = zipfile.ZipFile(new_directory + "/IMC_data.zip")
    zip_archive.extractall(new_directory)

## Remove, probably -- this previously coded for a widget on the opening screen that was an alternate way of inputting the directories for projects.
'''
class directory_display_helper_frame(ctk.CTkFrame):
    def __init__(self, master, dir_disp):
        super().__init__(master)
        self.master = master

        self.dir_disp = dir_disp

        label = ctk.CTkLabel(master = self, text = "Working Directory:")
        label.grid(column = 0, row = 0, pady = 3)

        self.directory = ctk.StringVar(value = self.dir_disp.currentdir)

        self.directory_entry = ctk.CTkEntry(master = self, textvariable = self.directory, width = 360)
        self.directory_entry.grid(column = 0, columnspan = 2, row = 1, pady = 3)

        self.dir_disp.bind("<Enter>", lambda enter:  self.get_current_directory())  
            ## helpful video that helped get started with events: 
            #       https://www.youtube.com/watch?v=8mZ9lZlsDHY&list=PLpMixYKO4EXeaGnqT_YWx7_mA77bz2VqM&index=9

        button2 = ctk.CTkButton(master = self, text = "Save working directory as default", command = self.set_as_home_directory)
        button2.grid(column = 1, row = 3, pady = 3)

        label2 = ctk.CTkLabel(master = self, text = "Launch from Directory Display Options:")
        label2.grid(column = 0, row = 4, columnspan = 2, pady = 3)

        self.frame = ctk.CTkFrame(master = self)
        self.frame.grid(column = 0, row = 5, columnspan = 2, pady = 3)

        self.launch_selection = ctk.StringVar(value = "mcd")
        radio1 = ctk.CTkRadioButton(master = self.frame, 
                                    text = "from MCD files", 
                                    variable = self.launch_selection, 
                                    value = "mcd")
        radio1.grid(row = 0, column = 0, padx = 1, pady = 1)

        radio2 = ctk.CTkRadioButton(master = self.frame, 
                                    text = "from .tiff files", 
                                    variable = self.launch_selection, 
                                    value = "tiff")
        radio2.grid(row = 0, column = 1, padx = 1, pady = 1)

        radio3 = ctk.CTkRadioButton(master = self.frame, 
                                    text = "from .fcs files", 
                                    variable = self.launch_selection, 
                                    value = "fcs")
        radio3.grid(row = 0, column = 2, padx = 1, pady = 1)

        button3 = ctk.CTkButton(master = self, text = "Launch Analysis from working directory", command = self.entry_from_dir_disp)
        button3.grid(column = 1, row = 6, pady = 3) 

    def entry_from_dir_disp(self) -> None:
        directory = self.directory_entry.get().strip()
        launch_selection = self.launch_selection.get()
        if launch_selection == "mcd":
            self.master.img_entry_func(directory = directory, from_mcds = True)
        elif launch_selection == "tiff":
            self.master.img_entry_func(directory = directory, from_mcds = False)
        elif launch_selection == "fcs":
            self.master.FCS_choice(directory = directory)

    def get_current_directory(self) -> None:
        directory = self.dir_disp.currentdir
        self.directory.set(directory)

    def set_as_home_directory(self) -> None:
        directory = self.directory.get()
        if directory != "None":
            try:
                os.listdir(directory)
            except FileNotFoundError:
                tk.messagebox.showwarning("Warning!", message = "This is not a valid directory to save!")
                return
        with open(chosen_dir_file, mode = 'w') as file:
            file.write(directory)
        self.dir_disp.setup_with_dir(directory)
'''



'''
This is a "vendorized" version of readimc package: https://github.com/BodenmillerGroup/readimc

(
MIT License

Copyright (c) 2021 University of Zurich

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
)

Changes:
    -- Kept only code needed for the MCDFile class of readimc  (the only part used by PalmettoBUG)
    -- unified code into one file (this file!) -- removing redundant / unused imports as they arose
    -- removed typing hints for Annotation, Panorama, Slide 
    -- switch to defusedxml for xml parsing and removed ET.Element type hinting
    -- add __all__ for docs
'''
__all__ = []
from warnings import warn
import mmap
import itertools
import re
from abc import ABC, abstractmethod
from os import PathLike
from typing import BinaryIO, List, Optional, Union, Dict, Tuple, Sequence
# from warnings import warn
import math
from dataclasses import dataclass, field

from pathlib import Path
import numpy as np
from imageio.v2 import imread

from xml.etree import ElementTree as ET


class AcquisitionBase(ABC):
    """Shared IMC acquisition metadata interface"""

    @property
    @abstractmethod
    def num_channels(self) -> int:
        """Number of channels"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def channel_metals(self) -> Sequence[str]:
        """Symbols of metal isotopes (e.g. ``["Ag", "Ir"]``)"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def channel_masses(self) -> Sequence[int]:
        """Atomic masses of metal isotopes (e.g. ``[107, 191]``)"""
        raise NotImplementedError()

    @property
    @abstractmethod
    def channel_labels(self) -> Sequence[str]:
        """Channel labels (user-provided)"""
        raise NotImplementedError()

    @property
    def channel_names(self) -> Sequence[str]:
        """Unique channel names in the format ``f"{metal}{mass}"`` (e.g.
        ``["Ag107", "Ir191"]``)"""
        return [
            f"{channel_metal}{channel_mass}"
            for channel_metal, channel_mass in zip(
                self.channel_metals, self.channel_masses
            )
        ]


@dataclass
class Acquisition(AcquisitionBase):
    """IMC acquisition metadata"""

    slide: "Slide"
    """Parent slide"""

    panorama: Optional["Panorama"]
    """Associated panorama"""

    id: int
    """Acquisition ID"""

    roi_points_um: Optional[
        Tuple[
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
        ]
    ]
    """User-provided ROI points, in micrometers

    Order: (top left, top right, bottom right, bottom left)"""

    metadata: Dict[str, str]
    """Full acquisition metadata"""

    _num_channels: int
    _channel_metals: List[str] = field(default_factory=list)
    _channel_masses: List[int] = field(default_factory=list)
    _channel_labels: List[str] = field(default_factory=list)

    @property
    def description(self) -> Optional[str]:
        """User-provided acquisition description"""
        return self.metadata.get("Description")

    @property
    def width_px(self) -> Optional[int]:
        """Acquisition width, in pixels"""
        value = self.metadata.get("MaxX")
        if value is not None:
            return int(value)
        return None

    @property
    def height_px(self) -> Optional[int]:
        """Acquisition height, in pixels"""
        value = self.metadata.get("MaxY")
        if value is not None:
            return int(value)
        return None

    @property
    def pixel_size_x_um(self) -> Optional[float]:
        """Width of a single pixel, in micrometers"""
        value = self.metadata.get("AblationDistanceBetweenShotsX")
        if value is not None:
            return float(value)
        return None

    @property
    def pixel_size_y_um(self) -> Optional[float]:
        """Height of a single pixel, in micrometers"""
        value = self.metadata.get("AblationDistanceBetweenShotsY")
        if value is not None:
            return float(value)
        return None

    @property
    def width_um(self) -> Optional[float]:
        """Acquisition width, in micrometers"""
        if self.width_px is not None and self.pixel_size_x_um is not None:
            return self.width_px * self.pixel_size_x_um
        return None

    @property
    def height_um(self) -> Optional[float]:
        """Acquisition height, in micrometers"""
        if self.height_px is not None and self.pixel_size_y_um is not None:
            return self.height_px * self.pixel_size_y_um
        return None

    @property
    def num_channels(self) -> int:
        return self._num_channels

    @property
    def channel_metals(self) -> Sequence[str]:
        return self._channel_metals

    @property
    def channel_masses(self) -> Sequence[int]:
        return self._channel_masses

    @property
    def channel_labels(self) -> Sequence[str]:
        return self._channel_labels

    @property
    def roi_coords_um(
        self,
    ) -> Optional[
        Tuple[
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
        ]
    ]:
        """ROI stage coordinates, in micrometers

        Order: (top left, top right, bottom right, bottom left)"""
        x1_str = self.metadata.get("ROIStartXPosUm")
        y1_str = self.metadata.get("ROIStartYPosUm")
        x3_str = self.metadata.get("ROIEndXPosUm")
        y3_str = self.metadata.get("ROIEndYPosUm")
        if (
            x1_str != x3_str
            and y1_str != y3_str
            and x1_str is not None
            and y1_str is not None
            and x3_str is not None
            and y3_str is not None
            and self.width_um is not None
            and self.height_um is not None
        ):
            x1, y1 = float(x1_str), float(y1_str)
            x3, y3 = float(x3_str), float(y3_str)
            # fix Fluidigm bug, where start positions are multiplied by 1000
            if abs(x1 / 1000.0 - x3) < abs(x1 - x3):
                x1 /= 1000.0
            if abs(y1 / 1000.0 - y3) < abs(y1 - y3):
                y1 /= 1000.0
            # calculate counter-clockwise rotation angle, in radians
            rotated_main_diag_angle = np.arctan2(y1 - y3, x1 - x3)
            main_diag_angle = np.arctan2(self.height_um, -self.width_um)
            angle = rotated_main_diag_angle - main_diag_angle
            # calculate missing points (generative approach)
            x2, y2 = self.width_um / 2.0, self.height_um / 2.0
            x4, y4 = -self.width_um / 2.0, -self.height_um / 2.0
            x2, y2 = (
                math.cos(angle) * x2 - math.sin(angle) * y2 + (x1 + x3) / 2.0,
                math.sin(angle) * x2 + math.cos(angle) * y2 + (y1 + y3) / 2.0,
            )
            x4, y4 = (
                math.cos(angle) * x4 - math.sin(angle) * y4 + (x1 + x3) / 2.0,
                math.sin(angle) * x4 + math.cos(angle) * y4 + (y1 + y3) / 2.0,
            )
            return ((x1, y1), (x2, y2), (x3, y3), (x4, y4))
        return None

@dataclass
class Panorama:
    """Panorama metadata (only for panoramas with panorama image data)"""

    slide: "Slide"
    """Parent slide"""

    id: int
    """Panorama ID"""

    metadata: Dict[str, str]
    """Full panorama metadata"""

    acquisitions: List[Acquisition] = field(default_factory=list)
    """List of acquisitions associated with this panorama"""

    @property
    def description(self) -> Optional[str]:
        """User-provided panorama description"""
        return self.metadata.get("Description")

    @property
    def width_um(self) -> Optional[float]:
        """Panorama width, in micrometers"""
        if self.points_um is not None:
            (x1, y1), (x2, y2), (x3, y3), (x4, y4) = self.points_um
            w1 = ((x1 - x2) ** 2.0 + (y1 - y2) ** 2.0) ** 0.5
            w2 = ((x3 - x4) ** 2.0 + (y3 - y4) ** 2.0) ** 0.5
            if abs(w1 - w2) > 0.001:
                raise ValueError(f"Panorama {self.id}: inconsistent image widths")
            return (w1 + w2) / 2.0
        return None

    @property
    def height_um(self) -> Optional[float]:
        """Panorama height, in micrometers"""
        if self.points_um is not None:
            (x1, y1), (x2, y2), (x3, y3), (x4, y4) = self.points_um
            h1 = ((x1 - x4) ** 2.0 + (y1 - y4) ** 2.0) ** 0.5
            h2 = ((x2 - x3) ** 2.0 + (y2 - y3) ** 2.0) ** 0.5
            if abs(h1 - h2) > 0.001:
                raise ValueError(f"Panorama {self.id}: inconsistent image heights")
            return (h1 + h2) / 2.0
        return None

    @property
    def points_um(
        self,
    ) -> Optional[
        Tuple[
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
            Tuple[float, float],
        ]
    ]:
        """User-provided ROI points, in micrometers

        Order: (top left, top right, bottom right, bottom left)"""
        x1_str = self.metadata.get("SlideX1PosUm")
        y1_str = self.metadata.get("SlideY1PosUm")
        x2_str = self.metadata.get("SlideX2PosUm")
        y2_str = self.metadata.get("SlideY2PosUm")
        x3_str = self.metadata.get("SlideX3PosUm")
        y3_str = self.metadata.get("SlideY3PosUm")
        x4_str = self.metadata.get("SlideX4PosUm")
        y4_str = self.metadata.get("SlideY4PosUm")
        if (
            x1_str is not None
            and y1_str is not None
            and x2_str is not None
            and y2_str is not None
            and x3_str is not None
            and y3_str is not None
            and x4_str is not None
            and y4_str is not None
        ):
            return (
                (float(x1_str), float(y1_str)),
                (float(x2_str), float(y2_str)),
                (float(x3_str), float(y3_str)),
                (float(x4_str), float(y4_str)),
            )
        return None

@dataclass
class Slide:
    """Slide metadata"""

    id: int
    """Slide ID"""

    metadata: Dict[str, str]
    """Full slide metadata"""

    panoramas: List[Panorama] = field(default_factory=list)
    """List of panoramas associated with this slide"""

    acquisitions: List[Acquisition] = field(default_factory=list)
    """List of acquisitions associated with this slide"""

    @property
    def description(self) -> Optional[str]:
        """User-provided slide description"""
        return self.metadata.get("Description")

    @property
    def width_um(self) -> Optional[float]:
        """Slide width, in micrometers"""
        value = self.metadata.get("WidthUm")
        if value is not None:
            return float(value)
        return None

    @property
    def height_um(self) -> Optional[float]:
        """Slide height, in micrometers"""
        value = self.metadata.get("HeightUm")
        if value is not None:
            return float(value)
        return None

class IMCFile(ABC):
    """Shared IMC file interface"""

    def __init__(self, path: Union[str, PathLike]) -> None:
        super().__init__()
        self._path = Path(path)

    @property
    def path(self) -> Path:
        """Path to the IMC file"""
        return self._path

    @abstractmethod
    def read_acquisition(self, acquisition = None) -> np.ndarray:
        """Reads IMC acquisition data as numpy array.

        :param acquisition: the acquisition to read
        :return: the acquisition data as 32-bit floating point array,
            shape: (c, y, x)
        """
        raise NotImplementedError()

class MCDParserError(Exception):
    def __init__(self, *args) -> None:
        """Error occurring when parsing invalid IMC .mcd file metadata"""
        super(MCDParserError, self).__init__(*args)


class MCDParser:
    _XMLNS_REGEX = re.compile(r"{(?P<xmlns>.*)}")
    _CHANNEL_REGEX = re.compile(r"^(?P<metal>[a-zA-Z]+)\((?P<mass>[0-9]+)\)$")

    def __init__(self, schema_xml: str) -> None:
        """A class for parsing IMC .mcd file metadata

        :param schema_xml: IMC .mcd file metadata in proprietary XML format
        """
        self._schema_xml = schema_xml
        self._schema_xml_elem = ET.fromstring(self._schema_xml)
        m = self._XMLNS_REGEX.match(self._schema_xml_elem.tag)
        self._schema_xml_xmlns = m.group("xmlns") if m is not None else None

    @property
    def schema_xml(self) -> str:
        """Full IMC .mcd file metadata in proprietary XML format"""
        return self._schema_xml

    @property
    def schema_xml_elem(self):
        """Full IMC .mcd file metadata as Python ElementTree element"""
        return self._schema_xml_elem

    @property
    def schema_xml_xmlns(self) -> Optional[str]:
        """Value of the metadata `xmlns` XML namespace attribute"""
        return self._schema_xml_xmlns

    @property
    def metadata(self) -> str:
        """Legacy accessor for `schema_xml`"""
        warn(
            "`MCDParser.metadata` will be removed in future readimc releases; "
            "use `MCDFile.schema_xml` instead"
        )
        return self.schema_xml

    @property
    def metadata_elem(self):
        """Legacy accessor for `schema_xml_elem`"""
        warn(
            "`MCDParser.metadata_elem` will be removed in future readimc releases; "
            "use `MCDFile.schema_xml_elem` instead"
        )
        return self.schema_xml_elem

    @property
    def metadata_xmlns(self) -> Optional[str]:
        """Legacy accessor for `schema_xml_xmlns`"""
        warn(
            "`MCDParser.metadata_xmlns` will be removed in future readimc releases; "
            "use `MCDFile.schema_xml_xmlns` instead"
        )
        return self.schema_xml_xmlns

    def parse_slides(self):
        """Extract slide metadata"""
        slides = [
            self._parse_slide(slide_elem) for slide_elem in self._find_elements("Slide")
        ]
        slides.sort(key=lambda slide: slide.id)
        return slides

    def _parse_slide(self, slide_elem):
        slide = Slide(
            self._get_text_as_int(slide_elem, "ID"),
            self._get_metadata_dict(slide_elem),
        )
        panorama_elems = self._find_elements(f"Panorama[SlideID='{slide.id}']")
        for panorama_elem in panorama_elems:
            panorama = None
            panorama_id = self._get_text_as_int(panorama_elem, "ID")
            panorama_type = self._get_text_or_none(panorama_elem, "Type")
            if panorama_type != "Default":  # ignore "virtual" Panoramas
                panorama = self._parse_panorama(panorama_elem, slide)
                slide.panoramas.append(panorama)
            acquisition_roi_elems = self._find_elements(
                f"AcquisitionROI[PanoramaID='{panorama_id}']"
            )
            for acquisition_roi_elem in acquisition_roi_elems:
                acquisition_roi_id = self._get_text_as_int(acquisition_roi_elem, "ID")
                roi_point_elems = self._find_elements(
                    f"ROIPoint[AcquisitionROIID='{acquisition_roi_id}']"
                )
                roi_points_um = None
                if len(roi_point_elems) == 4:
                    roi_points_um = tuple(
                        (
                            self._get_text_as_float(roi_point_elem, "SlideXPosUm"),
                            self._get_text_as_float(roi_point_elem, "SlideYPosUm"),
                        )
                        for roi_point_elem in sorted(
                            roi_point_elems,
                            key=lambda roi_point_elem: self._get_text_as_int(
                                roi_point_elem, "OrderNumber"
                            ),
                        )
                    )
                acquisition_elems = self._find_elements(
                    f"Acquisition[AcquisitionROIID='{acquisition_roi_id}']"
                )
                for acquisition_elem in acquisition_elems:
                    acquisition = self._parse_acquisition(
                        acquisition_elem, slide, panorama, roi_points_um  # type: ignore
                    )
                    slide.acquisitions.append(acquisition)
                    if panorama is not None:
                        panorama.acquisitions.append(acquisition)
        for a, b in itertools.combinations(slide.acquisitions, 2):
            a_start = int(a.metadata["DataStartOffset"])
            a_end = int(a.metadata["DataEndOffset"])
            b_start = int(b.metadata["DataStartOffset"])
            b_end = int(b.metadata["DataEndOffset"])
            if b_start <= a_start < b_end or b_start < a_end <= b_end:
                warn(
                    f"Slide {slide.id} corrupted: "
                    f"overlapping memory blocks for acquisitions {a.id} and {b.id}"
                )
        slide.panoramas.sort(key=lambda panorama: panorama.id)
        slide.acquisitions.sort(key=lambda acquisition: acquisition.id)
        return slide

    def _parse_panorama(self, panorama_elem, slide):
        return Panorama(
            slide,
            self._get_text_as_int(panorama_elem, "ID"),
            self._get_metadata_dict(panorama_elem),
        )

    def _parse_acquisition(
        self,
        acquisition_elem,
        slide,
        panorama,
        roi_points_um: Optional[
            Tuple[
                Tuple[float, float],
                Tuple[float, float],
                Tuple[float, float],
                Tuple[float, float],
            ]
        ],
    ):
        acquisition_id = self._get_text_as_int(acquisition_elem, "ID")
        acquisition_channel_elems = self._find_elements(
            f"AcquisitionChannel[AcquisitionID='{acquisition_id}']"
        )
        acquisition_channel_elems.sort(
            key=lambda acquisition_channel_elem: self._get_text_as_int(
                acquisition_channel_elem, "OrderNumber"
            )
        )
        acquisition = Acquisition(
            slide,
            panorama,
            acquisition_id,
            roi_points_um,
            self._get_metadata_dict(acquisition_elem),
            len(acquisition_channel_elems) - 3,
        )
        for i, acquisition_channel_elem in enumerate(acquisition_channel_elems):
            channel_name = self._get_text(acquisition_channel_elem, "ChannelName")
            if i == 0 and channel_name != "X":
                raise MCDParserError(
                    f"First channel '{channel_name}' should be named 'X'"
                )
            if i == 1 and channel_name != "Y":
                raise MCDParserError(
                    f"Second channel '{channel_name}' should be named 'Y'"
                )
            if i == 2 and channel_name != "Z":
                raise MCDParserError(
                    f"Third channel '{channel_name}' should be named 'Z'"
                )
            if channel_name in ("X", "Y", "Z"):
                continue
            m = self._CHANNEL_REGEX.match(channel_name)
            if m is None:
                raise MCDParserError(
                    "Cannot extract channel information "
                    f"from channel name '{channel_name}' "
                    f"for acquisition {acquisition.id}"
                )
            channel_label = self._get_text(acquisition_channel_elem, "ChannelLabel")
            acquisition._channel_metals.append(m.group("metal"))
            acquisition._channel_masses.append(int(m.group("mass")))
            acquisition._channel_labels.append(channel_label)
        return acquisition

    def _find_elements(self, path: str):
        namespaces = None
        if self._schema_xml_xmlns is not None:
            namespaces = {"": self._schema_xml_xmlns}
        return self._schema_xml_elem.findall(path, namespaces=namespaces)

    def _get_text_or_none(self, parent_elem, tag: str) -> Optional[str]:
        namespaces = None
        if self._schema_xml_xmlns is not None:
            namespaces = {"": self._schema_xml_xmlns}
        elem = parent_elem.find(tag, namespaces=namespaces)
        return (elem.text or "") if elem is not None else None

    def _get_text(self, parent_elem, tag: str) -> str:
        text = self._get_text_or_none(parent_elem, tag)
        if text is None:
            raise MCDParserError(
                f"XML tag '{tag}' not found for parent XML tag '{parent_elem.tag}'"
            )
        return text

    def _get_text_as_int(self, parent_elem, tag: str) -> int:
        text = self._get_text(parent_elem, tag)
        try:
            return int(text)
        except ValueError as e:
            raise MCDParserError(
                f"Text '{text}' of XML tag '{tag}' cannot be converted to int "
                f"for parent XML tag '{parent_elem.tag}'"
            ) from e

    def _get_text_as_float(self, parent_elem, tag: str) -> float:
        text = self._get_text(parent_elem, tag)
        try:
            return float(text)
        except ValueError as e:
            raise MCDParserError(
                f"Text '{text}' of XML tag '{tag}' cannot be converted to "
                f"float for parent XML tag '{parent_elem.tag}'"
            ) from e

    def _get_metadata_dict(self, parent_elem) -> Dict[str, str]:
        metadata = {}
        for elem in parent_elem:
            tag = elem.tag
            if self._schema_xml_xmlns is not None:
                tag = tag.replace(f"{{{self._schema_xml_xmlns}}}", "")
            metadata[tag] = elem.text or ""
        return metadata

class MCDFile(IMCFile):
    def __init__(self, path: Union[str, PathLike]) -> None:
        """A class for reading IMC .mcd files

        :param path: path to the IMC .mcd file
        """
        super(MCDFile, self).__init__(path)
        self._fh: Optional[BinaryIO] = None
        self._schema_xml: Optional[str] = None
        self._slides = None

    @property
    def schema_xml(self) -> str:
        """Full metadata in proprietary XML format"""
        if self._schema_xml is None:
            raise IOError(f"MCD file '{self.path.name}' has not been opened")
        return self._schema_xml

    @property
    def metadata(self) -> str:
        """Legacy accessor for `schema_xml`"""
        warn(
            "`MCDFile.metadata` will be removed in future readimc releases; "
            "use `MCDFile.schema_xml` instead"
        )
        return self.schema_xml

    @property
    def slides(self):
        """Metadata on slides contained in this IMC .mcd file"""
        if self._slides is None:
            raise IOError(f"MCD file '{self.path.name}' has not been opened")
        return self._slides

    def __enter__(self) -> "MCDFile":
        self.open()
        return self

    def __exit__(self, exc_type, exc_value, traceback) -> None:
        self.close()

    def open(self) -> None:
        """Opens the IMC .mcd file for reading.

        It is good practice to use context managers whenever possible:

        .. code-block:: python

            with MCDFile("/path/to/file.mcd") as f:
                pass

        """
        if self._fh is not None:
            self._fh.close()
        self._fh = open(self._path, mode="rb")
        self._schema_xml = self._read_schema_xml()
        try:
            self._slides = MCDParser(self._schema_xml).parse_slides()
        except MCDParserError as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                "error parsing slide information from MCD-XML"
            ) from e

    def close(self) -> None:
        """Closes the IMC .mcd file.

        It is good practice to use context managers whenever possible:

        .. code-block:: python

            with MCDFile("/path/to/file.mcd") as f:
                pass

        """
        if self._fh is not None:
            self._fh.close()
            self._fh = None

    def read_acquisition(
        self, acquisition = None, strict: bool = True
    ) -> np.ndarray:
        """Reads IMC acquisition data as numpy array.

        :param acquisition: the acquisition to read
        :param strict: set this parameter to False to try to recover corrupted data
        :return: the acquisition data as 32-bit floating point array,
            shape: (c, y, x)
        """
        if acquisition is None:
            raise ValueError("acquisition")
        if self._fh is None:
            raise IOError(f"MCD file '{self.path.name}' has not been opened")
        try:
            data_start_offset = int(acquisition.metadata["DataStartOffset"])
            data_end_offset = int(acquisition.metadata["DataEndOffset"])
            value_bytes = int(acquisition.metadata["ValueBytes"])
        except (KeyError, ValueError) as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                "cannot locate acquisition image data"
            ) from e
        if data_start_offset >= data_end_offset:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                "invalid acquisition image data offsets"
            )
        if value_bytes <= 0:
            raise IOError("MCD file corrupted: invalid byte size")
        num_channels = acquisition.num_channels
        data_size = data_end_offset - data_start_offset
        bytes_per_pixel = (num_channels + 3) * value_bytes
        if data_size % bytes_per_pixel != 0:
            data_size += 1
        if data_size % bytes_per_pixel != 0:
            if strict:
                raise IOError(
                    f"MCD file '{self.path.name}' corrupted: "
                    "invalid acquisition image data size"
                )
            warn(
                f"MCD file '{self.path.name}' corrupted: "
                "invalid acquisition image data size"
            )
        num_pixels = data_size // bytes_per_pixel
        self._fh.seek(0)
        data = np.memmap(
            self._fh,
            dtype=np.float32,
            mode="r",
            offset=data_start_offset,
            shape=(num_pixels, num_channels + 3),
        )
        xs = data[:, 0].astype(int)
        ys = data[:, 1].astype(int)
        try:
            width = int(acquisition.metadata["MaxX"])
            height = int(acquisition.metadata["MaxY"])
            if width <= np.amax(xs) or height <= np.amax(ys):
                raise ValueError(
                    "data shape is incompatible with acquisition image dimensions"
                )
        except (KeyError, ValueError):
            warn(
                f"MCD file '{self.path.name}' corrupted: "
                "cannot read acquisition image dimensions; recovering from data shape"
            )
            width = np.amax(xs) + 1
            height = np.amax(ys) + 1
        if width * height != data.shape[0]:
            if strict:
                raise IOError(
                    f"MCD file '{self.path.name}' corrupted: "
                    "inconsistent acquisition image data size"
                )
            warn(
                f"MCD file '{self.path.name}' corrupted: "
                "inconsistent acquisition image data size"
            )
        img = np.zeros((num_channels, height, width), dtype=np.float32)
        img[:, ys, xs] = np.transpose(data[:, 3:])
        return img

    def read_slide(
        self, slide, raw: bool = False
    ) -> Union[np.ndarray, bytes, None]:
        """Reads and decodes a slide image as numpy array using the ``imageio``
        package.

        .. note::
            Slide images are stored as binary data within the IMC .mcd file in
            an arbitrary encoding. The ``imageio`` package can decode most
            commonly used image file formats, but may fail for more obscure,
            in which case an ``IOException`` is raised.

        :param slide: the slide to read
        :return: the slide image, or ``None`` if no image is available for the
            specified slide
        """
        try:
            data_start_offset = int(slide.metadata["ImageStartOffset"])
            data_end_offset = int(slide.metadata["ImageEndOffset"])
        except (KeyError, ValueError) as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot locate image data for slide {slide.id}"
            ) from e
        if data_start_offset == data_end_offset == 0:
            return None
        data_start_offset += 161
        data_end_offset -= 1
        if data_start_offset >= data_end_offset:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"invalid image data offsets for slide {slide.id}"
            )
        try:
            return self._read_image(
                data_start_offset, data_end_offset - data_start_offset, raw
            )
        except Exception as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot read image for slide {slide.id}"
            ) from e

    def read_panorama(
        self, panorama, raw: bool = False
    ) -> Union[np.ndarray, bytes, None]:
        """Reads and decodes a panorama image as numpy array using the
        ``imageio`` package.

        :param panorama: the panorama to read
        :return: the panorama image as numpy array
        """
        try:
            data_start_offset = int(panorama.metadata["ImageStartOffset"])
            data_end_offset = int(panorama.metadata["ImageEndOffset"])
        except (KeyError, ValueError) as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot locate image data for panorama {panorama.id}"
            ) from e
        if data_start_offset == data_end_offset == 0:
            return None
        data_start_offset += 161
        data_end_offset -= 1
        if data_start_offset >= data_end_offset:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"invalid image data offsets for panorama {panorama.id}"
            )
        try:
            return self._read_image(
                data_start_offset, data_end_offset - data_start_offset, raw
            )
        except Exception as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot read image for panorama {panorama.id}"
            ) from e

    def read_before_ablation_image(
        self, acquisition, raw: bool = False
    ) -> Union[np.ndarray, bytes, None]:
        """Reads and decodes a before-ablation image as numpy array using the
        ``imageio`` package.

        :param acquisition: the acquisition for which to read the
            before-ablation image
        :return: the before-ablation image as numpy array, or ``None`` if no
            before-ablation image is available for the specified acquisition
        """
        try:
            data_start_offset = int(
                acquisition.metadata["BeforeAblationImageStartOffset"]
            )
            data_end_offset = int(acquisition.metadata["BeforeAblationImageEndOffset"])
        except (KeyError, ValueError) as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot locate before-ablation image data "
                f"for acquisition {acquisition.id}"
            ) from e
        if data_start_offset == data_end_offset == 0:
            return None
        data_start_offset += 161
        data_end_offset -= 1
        if data_start_offset >= data_end_offset:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"invalid before-ablation image data offsets "
                f"for acquisition {acquisition.id}"
            )
        try:
            return self._read_image(
                data_start_offset, data_end_offset - data_start_offset, raw
            )
        except Exception as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot read before-ablation image "
                f"for acquisition {acquisition.id}"
            ) from e

    def read_after_ablation_image(
        self, acquisition, raw: bool = False
    ) -> Union[np.ndarray, bytes, None]:
        """Reads and decodes a after-ablation image as numpy array using the
        ``imageio`` package.

        :param acquisition: the acquisition for which to read the
            after-ablation image
        :return: the after-ablation image as numpy array, or ``None`` if no
            after-ablation image is available for the specified acquisition
        """
        try:
            data_start_offset = int(
                acquisition.metadata["AfterAblationImageStartOffset"]
            )
            data_end_offset = int(acquisition.metadata["AfterAblationImageEndOffset"])
        except (KeyError, ValueError) as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot locate after-ablation image data "
                f"for acquisition {acquisition.id}"
            ) from e
        if data_start_offset == data_end_offset == 0:
            return None
        data_start_offset += 161
        data_end_offset -= 1
        if data_start_offset >= data_end_offset:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"invalid after-ablation image data offsets "
                f"for acquisition {acquisition.id}"
            )
        try:
            return self._read_image(
                data_start_offset, data_end_offset - data_start_offset, raw
            )
        except Exception as e:
            raise IOError(
                f"MCD file '{self.path.name}' corrupted: "
                f"cannot read after-ablation image "
                f"for acquisition {acquisition.id}"
            ) from e

    def _read_schema_xml(
        self,
        encoding: str = "utf-16-le",
        start_sub: str = "<MCDSchema",
        end_sub: str = "</MCDSchema>",
    ) -> str:
        if self._fh is None:
            raise IOError(f"MCD file '{self.path.name}' has not been opened")
        with mmap.mmap(self._fh.fileno(), 0, access=mmap.ACCESS_READ) as mm:
            # V1 contains multiple MCDSchema entries
            # As per imctools, the latest entry should be taken
            start_sub_encoded = start_sub.encode(encoding=encoding)
            start_index = mm.rfind(start_sub_encoded)
            if start_index == -1:
                raise IOError(
                    f"MCD file '{self.path.name}' corrupted: "
                    f"start of XML document '{start_sub}' not found"
                )
            end_sub_encoded = end_sub.encode(encoding=encoding)
            end_index = mm.rfind(end_sub_encoded, start_index)
            if end_index == -1:
                raise IOError(
                    f"MCD file '{self.path.name}' corrupted: "
                    f"end of XML document '{end_sub}' not found"
                )
            mm.seek(start_index)
            data = mm.read(end_index + len(end_sub_encoded) - start_index)
        return data.decode(encoding=encoding)

    def _read_image(
        self, data_offset: int, data_size: int, raw: bool = False
    ) -> Union[np.ndarray, bytes]:
        if self._fh is None:
            raise IOError(f"MCD file '{self.path.name}' has not been opened")
        self._fh.seek(data_offset)
        data = self._fh.read(data_size)
        if raw:
            print('imread dependency is not used in this case from readimc -- consider removing!')
            return data
        else:
            print('imread dependency is used from readimc')
            return imread(data)

    def __repr__(self) -> str:
        return str(self._path)



# ruff: noqa
'''
This file is an 'vendorizing' of the fcsparser package: https://github.com/eyurtsev/fcsparser

(
The MIT License (MIT)

Copyright (c) 2018 Eugene Yurtsev

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
)

Changes:
    -- add __all__ (for the sake of autoapi docs -- I don't really want to include any of the vendor files)
    -- removed logging (replacing with print statement)
    -- removed Python 2 support (six package code)
    -- removed now unused imports (division, logging, six)


'''
__all__ = []
###### Begin copied code:


"""Parser for FCS 2.0, 3.0, 3.1 files. Python 2/3 compatible.

Distributed under the MIT License.

Useful documentation for dtypes in numpy
http://docs.scipy.org/doc/numpy/reference/generated/numpy.ndarray.byteswap.html?highlight=byteswap#numpy.ndarray.byteswap
http://docs.scipy.org/doc/numpy/user/basics.types.html
http://docs.scipy.org/doc/numpy/reference/arrays.dtypes.html
"""
import contextlib
import string
import sys
import warnings
from io import BytesIO

import numpy
import pandas as pd

def fromfile(file, dtype, count, *args, **kwargs):
    """Wrapper around np.fromfile to support any file-like object."""

    # to understand this function, you must understand the difference
    # between a numpy *record* and a numpy *field*.  a *record* is
    # composed of a number of *fields*, each of which may be a different
    # dtype.  for example, if we were encoding a list of two-dimensional
    # points on a plane, each point would be represented by a record
    # containing two fields, each of which was a four-byte float
    # represented by the dtype 'f4'.  thus, the record would be
    # specified by dtype("f4, f4").

    # numpy's functions expect all field widths to be a power of two
    # (because this is how such things are stored in memory.)
    # unfortunately, some pathological cytometers don't make their
    # records a power-of-two wide.  For example, a Cytek xP5 encodes
    # records in their DATA segment in three-byte-wide integers --
    # this comes in as a dtype of "i3", which makes numpy freak out.

    # To address this, we convert the requested dtype so that each
    # record is read as a series of one-byte-wide unsigned integers
    # ('u1'), pad out each record with NUL bytes until it is
    # a power-of-two wide, then re-convert it to the requested
    # dtype (but with a power-of-two width) and return it.

    # what dtypes were we asked for?
    dtypes = dtype.split(",")
    field_widths = []

    # how wide is each dtype?
    for dt in dtypes:
        num_bytes = int(dt[2:])
        field_widths.append(num_bytes)

    # how many bytes wide is the total record?
    record_width = sum(field_widths)

    # read the DATA segment into a 1 x `count` array of records.
    # each record has a number of `u1` (one-byte unsigned integers)
    # equal to `record_width`.
    try:
        ret = numpy.fromfile(
            file, dtype=",".join(["u1"] * record_width), count=count, *args, **kwargs
        )
    except (TypeError, IOError):
        _ret = numpy.frombuffer(
            file.read(count * record_width),
            dtype=",".join(["u1"] * record_width),
            count=count,
            *args,
            **kwargs
        )
        # Create a copy of the file content as `numpy.frombuffer`
        # returns a view into the original object which is not 
        # safe for mutable file buffers.
        # See https://numpy.org/doc/stable/reference/generated/numpy.frombuffer.html
        ret = _ret.copy()

    # convert the DATA segment from a 1 x `count` array of records
    # (and remember, each record is composed of `record_width`
    # 1-byte unsigned ints) to a `record_width` x `count` array of
    # 'u1' unsigned ints.
    ret = ret.view("u1").reshape((count, record_width))

    # now, for each requested dtype.....
    ret_dtypes = []
    for field_idx, dt in enumerate(dtypes):
        dtype_type = dt[1]
        dtype_endian = dt[0]
        num_bytes = int(dt[2:])

        # num_bytes & (num_bytes - 1) is 0 IFF num_bytes is a power of two
        # while num_bytes is NOT a power of two....
        while num_bytes & (num_bytes - 1) != 0:
            # ...insert another COLUMN of NUL bytes at the front of the field....
            ret = numpy.insert(
                ret, sum(field_widths[0:field_idx]), numpy.zeros(count), axis=1
            )

            # ....and increment the number of bytes for this field.
            num_bytes = num_bytes + 1

        # when we've got a field that's a power-of-two wide, append that field's
        # dtype to the list of record dtypes we're going to return
        ret_dtypes.append(dtype_endian + dtype_type + str(num_bytes))

    # now, "cast" the newly padded array as the desired data types,
    # and return it.
    return ret.view(",".join(ret_dtypes)).ravel()

class ParserFeatureNotImplementedError(Exception):
    """Raise when encountering fcs files for which parsing hasn't been implemented."""


class FCSParser(object):
    def __init__(
        self,
        path=None,
        read_data=True,
        channel_naming="$PnS",
        data_set=0,
        encoding="utf-8",
    ):
        """Parse FCS files.

        Compatible with most FCS 2.0, 3.0, 3.1 files.

        self.annotation: a dictionary holding the parsed content of the TEXT segment
                         In addition, a key called __header__ 
                         has been added to this dictionary
                         It specifies the information parsed 
                         from the FCS file HEADER segment.
                         (This won't be necessary for most users.)

        self.data holds the parsed DATA segment
        self.analysis holds the ANALYSIS segment as read from the file.

        After the data segment is read:
            self.channel_names holds the chosen names of the channels
            self.channel_names_alternate holds the alternate names of the channels

        Args:
            path : str
                Path of .fcs file
            read_data : bool
                If True, reads the data immediately.
                Otherwise, use read_data method to read in the data from the fcs file.
            channel_naming: '$PnS' | '$PnN'
                Determines which meta data field is used for naming the channels.

                The default should be $PnS 
                (even though it is not guaranteed to be unique)

                $PnN stands for the short name (guaranteed to be unique).
                    Will look like 'FL1-H'
                $PnS stands for the actual name (not guaranteed to be unique).
                    Will look like 'FSC-H' (Forward scatter)

                The chosen field will be used to population self.channels.

                If the chosen field does not exist in the fcs file.
                The program attempts to use the alternative field by default.

                Note: These names are not flipped in the implementation.
                It looks like they were swapped 
                for some reason in the official FCS specification.
            data_set: int
                Index of retrieved data set in the fcs file.
                This value specifies the data set being retrieved from an fcs file with
                multiple data sets.
            encoding: str
                Specify encoding of the text section of the fcs data
        """
        self._encoding = encoding
        self._data = None
        self._channel_naming = channel_naming
        self.channel_names_s = []
        self.channel_names_n = []

        # Attributes parsed from fcs file
        self._data_start = -1
        self._data_end = -1
        self.channel_numbers = []
        self._analysis = None
        self._file_size = 0

        if channel_naming not in ("$PnN", "$PnS"):
            raise ValueError('channel_naming must be either "$PnN" or "$PnS"')

        self.annotation = {}
        self.path = path

        if path:
            with open(path, "rb") as f:
                self.load_file(f, data_set=data_set, read_data=read_data)

    def load_file(self, file_handle, data_set=0, read_data=True):
        """Load the requested parts of the file into memory."""
        file_handle.seek(0, 2)
        self._file_size = file_handle.tell()
        file_handle.seek(0)
        data_segments = 0
        # seek the correct data set in fcs
        nextdata_offset = 0
        while data_segments <= data_set:
            self.read_header(file_handle, nextdata_offset)
            self.read_text(file_handle)
            if "$NEXTDATA" in self.annotation:
                data_segments += 1
                nextdata_offset += self.annotation["$NEXTDATA"]
                file_handle.seek(nextdata_offset)
                if nextdata_offset == 0 and data_segments < data_set:
                    warnings.warn("File does not contain the number of data sets.")
                    break
            else:
                if data_segments != 0:
                    warnings.warn("File does not contain $NEXTDATA information.")
                break
        if read_data:
            self.read_data(file_handle)

    @classmethod
    def from_data(cls, data):
        """Load an FCS file from a bytes-like object.

        Args:
            data: buffer containing contents of an FCS file.

        Returns:
            FCSParser instance with data loaded
        """
        obj = cls()
        with contextlib.closing(BytesIO(data)) as file_handle:
            obj.load_file(file_handle)
        return obj

    def read_header(self, file_handle, nextdata_offset=0):
        """Read the header of the FCS file.

        The header specifies where the annotation, data and analysis 
        are located inside the binary
        file.

        Args:
            file_handle: buffer containing FCS file.
            nextdata_offset: byte offset of a set header 
            from file start specified by $NEXTDATA
        """
        header = {"FCS format": file_handle.read(6)}

        file_handle.read(4)  # 4 space characters after the FCS format

        for field in (
            "text start",
            "text end",
            "data start",
            "data end",
            "analysis start",
            "analysis end",
        ):
            s = file_handle.read(8)
            try:
                field_value = int(s)
            except ValueError:
                field_value = 0
            header[field] = field_value + nextdata_offset
        # In some .fcs files, 'text end' and 'data start' are equal, e.g.,
        # http://flowrepository.org/experiments/2241/download_ziped_files
        # and this would lead to a mistake when run @_extract_text_dict
        # We should avoid this situation.
        if header["text end"] == header["data start"]:
            header["text end"] = header["text end"] - 1

        # Checking that the location of the TEXT segment is specified
        for k in ("text start", "text end"):
            if header[k] == 0:
                raise ValueError(
                    'The FCS file "{}" seems corrupted. (Parser cannot locate '
                    'information about the "{}" segment.)'.format(self.path, k)
                )
            elif header[k] > self._file_size:
                raise ValueError(
                    'The FCS file "{}" is corrupted. "{}" segment '
                    "is larger than file size".format(self.path, k)
                )
            else:
                # All OK
                pass

        self._data_start = header["data start"]
        self._data_end = header["data start"]

        if header["analysis end"] - header["analysis start"] != 0:
            warnings.warn(
                "There appears to be some information in the ANALYSIS segment of file "
                "{0}. However, it might not be read correctly.".format(self.path)
            )

        self.annotation["__header__"] = header

    @staticmethod
    def _extract_text_dict(raw_text):
        """Parse the TEXT segment of the FCS file into a python dictionary."""
        delimiter = raw_text[0]

        if raw_text[-1] != delimiter:
            # Avoid stripping whitespace delimiter
            if delimiter.strip() == delimiter:
                raw_text = raw_text.strip()
            if raw_text[-1] != delimiter:
                msg = (
                    "The first two characters were:\n {}. "
                    "The last two characters were: {}\n"
                    "Parser expects the same delimiter character in beginning "
                    "and end of TEXT segment. "
                    "This file may be parsed incorrectly!".format(
                        raw_text[:2], raw_text[-2:]
                    )
                )
                warnings.warn(msg)
                raw_text = raw_text[1:]
            else:
                raw_text = raw_text[1:-1]
        else:
            raw_text = raw_text[1:-1]

        # 1:-1 above removes the first and last characters which are reserved 
        # for the delimiter.

        # The delimiter is escaped by being repeated (two consecutive delimiters). 
        # This code splits on the escaped delimiter first, 
        # so there is no need for extra logic to distinguish
        # actual delimiters from escaped delimiters.
        nested_split_list = [x.split(delimiter) for x in raw_text.split(delimiter * 2)]

        # Flatten the nested list to a list of elements (alternating keys and values)
        raw_text_elements = nested_split_list[0]
        for partial_element_list in nested_split_list[1:]:
            # Rejoin two parts of an element that was split by an escaped delimiter 
            # (the end and start of two successive sub-lists in nested_split_list)
            raw_text_elements[-1] += delimiter + partial_element_list[0]
            raw_text_elements.extend(partial_element_list[1:])

        keys, values = raw_text_elements[0::2], raw_text_elements[1::2]
        return dict(zip(keys, values))

    def read_text(self, file_handle):
        """Parse the TEXT segment of the FCS file.

        The TEXT segment contains meta data associated with the FCS file.
        Converting all meta keywords to lower case.
        """
        header = self.annotation["__header__"]  # For convenience

        #####
        # Read in the TEXT segment of the FCS file
        # There are some differences in how the
        file_handle.seek(header["text start"], 0)
        raw_text = file_handle.read(header["text end"] - header["text start"] + 1)
        try:
            raw_text = raw_text.decode(self._encoding)
        except UnicodeDecodeError as e:
            # Catching the exception and logging it in this way kills the traceback, but
            # we can worry about this later.
            print("Encountered an illegal utf-8 byte in the header.\n Illegal utf-8 "
                "characters will be ignored.\n{}".format(e))
            raw_text = raw_text.decode(self._encoding, errors="ignore")

        text = self._extract_text_dict(raw_text)

        ##
        # Extract channel names and convert some of the channel properties
        # and other fields into numeric data types (from string)
        # Note: do not use regular expressions for manipulations here.
        # Regular expressions are too heavy in terms of computation time.
        pars = int(text["$PAR"])
        if (
            "$P0B" in text.keys()
        ):  # Checking whether channel number count starts from 0 or from 1
            self.channel_numbers = range(0, pars)  # Channel number count starts from 0
        else:
            self.channel_numbers = range(1, pars + 1)  # Channel numbers start from 1

        
        # Extract parameter names
        channel_names_n = []
        channel_names_s = []
        for channel_number in self.channel_numbers:
            n_key = f'$P{channel_number}N'
            s_key = f'$P{channel_number}S'
            name_n = text.get(n_key, "")
            name_s = text.get(s_key, "")
            
            channel_names_n.append(name_n)
            channel_names_s.append(name_s if name_s != "" else name_n)

        self.channel_names_n = tuple(channel_names_n)
        self.channel_names_s = tuple(channel_names_s)
        
        # Convert some of the fields into integer values
        keys_encoding_bits = [
            f"$P{channel_number}B" for channel_number in self.channel_numbers
        ]

        add_keys_to_convert_to_int = ["$NEXTDATA", "$PAR", "$TOT"]

        keys_to_convert_to_int = keys_encoding_bits + add_keys_to_convert_to_int

        for key in keys_to_convert_to_int:
            value = text[key]
            text[key] = int(value)

        self.annotation.update(text)

        # Update data start segments if needed

        if self._data_start == 0:
            self._data_start = int(text["$BEGINDATA"])
        if self._data_end == 0:
            self._data_end = int(text["$ENDDATA"])

    def read_analysis(self, file_handle):
        """Read the ANALYSIS segment of the FCS file and store it in self.analysis.

        Warning: This has never been tested with an actual fcs file that contains an
        analysis segment.

        Args:
            file_handle: buffer containing FCS data
        """
        start = self.annotation["__header__"]["analysis start"]
        end = self.annotation["__header__"]["analysis end"]
        if start != 0 and end != 0:
            file_handle.seek(start, 0)
            self._analysis = file_handle.read(end - start)
        else:
            self._analysis = None

    def _verify_assumptions(self):
        """Verify that all assumptions made by the parser hold."""
        text = self.annotation
        keys = text.keys()

        if "$MODE" not in text or text["$MODE"] != "L":
            raise ParserFeatureNotImplementedError("Mode not implemented")

        if "$P0B" in keys:
            raise ParserFeatureNotImplementedError(
                "Not expecting a parameter starting at 0"
            )

        if text["$BYTEORD"] not in ["1,2,3,4", "4,3,2,1", "1,2", "2,1"]:
            raise ParserFeatureNotImplementedError(
                "$BYTEORD {} " "not implemented".format(text["$BYTEORD"])
            )

    def get_channel_names(self):
        """Get list of channel names. Raises a warning if the names are not unique."""
        names_s, names_n = self.channel_names_s, self.channel_names_n

        # Figure out which channel names to use
        if self._channel_naming == "$PnS":
            channel_names, channel_names_alternate = names_s, names_n
        else:
            channel_names, channel_names_alternate = names_n, names_s

        if len(channel_names) == 0:
            channel_names = channel_names_alternate

        if len(set(channel_names)) != len(channel_names):
            msg = (
                "The default channel names (defined by the {} "
                "parameter in the FCS file) were not unique. To avoid "
                "problems in downstream analysis, the channel names "
                "have been switched to the alternate channel names "
                "defined in the FCS file. To avoid "
                "seeing this warning message, explicitly instruct "
                "the FCS parser to use the alternate channel names by "
                "specifying the channel_naming parameter."
            )
            msg = msg.format(self._channel_naming)
            warnings.warn(msg)
            channel_names = channel_names_alternate

        return channel_names

    def read_data(self, file_handle):
        """Read the DATA segment of the FCS file."""
        self._verify_assumptions()
        text = self.annotation

        if (self._data_start > self._file_size) or (self._data_end > self._file_size):
            raise ValueError(
                'The FCS file "{}" is corrupted. Part of the data segment '
                "is missing.".format(self.path)
            )

        num_events = text["$TOT"]  # Number of events recorded
        num_pars = text["$PAR"]  # Number of parameters recorded

        if text["$BYTEORD"].strip() == "1,2,3,4" or text["$BYTEORD"].strip() == "1,2":
            endian = "<"
        elif text["$BYTEORD"].strip() == "4,3,2,1" or text["$BYTEORD"].strip() == "2,1":
            endian = ">"
        else:
            msg = "Unrecognized byte order ({})".format(text["$BYTEORD"])
            raise ParserFeatureNotImplementedError(msg)

        # dictionary to convert from FCS format to numpy convention
        conversion_dict = {"F": "f", "D": "f", "I": "u"}

        if text["$DATATYPE"] not in conversion_dict.keys():
            raise ParserFeatureNotImplementedError(
                "$DATATYPE = {0} is not yet " "supported.".format(text["$DATATYPE"])
            )

        # Calculations to figure out data types of each of parameters
        # $PnB specifies the number of bits reserved for a measurement of parameter n
        bytes_per_par_list = [
            int(text["$P{0}B".format(i)] / 8) for i in self.channel_numbers
        ]

        par_numeric_type_list = [
            "{endian}{type}{size}".format(
                endian=endian,
                type=conversion_dict[text["$DATATYPE"]],
                size=bytes_per_par,
            )
            for bytes_per_par in bytes_per_par_list
        ]

        # Parser for list mode. Here, the order is a list of tuples.
        # Each tuple stores event related information
        file_handle.seek(
            self._data_start, 0
        )  # Go to the part of the file where data starts

        ##
        # Read in the data
        if len(set(par_numeric_type_list)) > 1:
            # This branch deals with files in which the different columns (channels)
            # were encoded with different types; i.e., a mixed data format.
            dtype = ",".join(par_numeric_type_list)
            data = fromfile(file_handle, dtype, num_events)

            # The dtypes in the numpy array `data` above are associated with both a name
            # and a type; i.e.,
            # https://docs.scipy.org/doc/numpy/reference/generated/numpy.recarray.html
            # The names are assigned automatically.
            # In order for this code to work correctly 
            # with the pandas DataFrame constructor,
            # we convert the *names* of the dtypes to the channel names we want to use.

            names = self.get_channel_names()

            encoded_names = [name for name in names]

            data.dtype.names = tuple(encoded_names)
        else:
            # values saved in a single data format
            dtype = par_numeric_type_list[0]
            data = fromfile(file_handle, dtype, num_events * num_pars)
            data = data.reshape((num_events, num_pars))
        ##
        # Convert to native byte order
        # This is needed for working with pandas data structures
        native_code = "<" if (sys.byteorder == "little") else ">"
        if endian != native_code:
            # swaps the actual bytes and also the endianness
            data = data.byteswap().newbyteorder()

        # Mask off high bits if integer type data
        if text["$DATATYPE"] == "I":
            if len(set(par_numeric_type_list)) > 1:
                for channel_number in self.channel_numbers:
                    valid_bits = numpy.ceil(
                        numpy.log2(float(text["$P{0}R".format(channel_number)]))
                    )

                    if bytes_per_par_list[channel_number - 1] * 8 == valid_bits:
                        continue

                    name = data.dtype.names[channel_number - 1]
                    bitmask = numpy.array([2**valid_bits - 1], dtype=data[name].dtype)
                    data[name] = data[name] & bitmask
            else:
                valid_bits_per_par_list = numpy.array(
                    [
                        2 ** numpy.ceil(numpy.log2(float(text["$P{0}R".format(i)]))) - 1
                        for i in self.channel_numbers
                    ],
                    dtype=data.dtype,
                )
                data &= valid_bits_per_par_list

        self._data = data

    @property
    def data(self):
        """Get parsed DATA segment of the FCS file."""
        if self._data is None:
            with open(self.path, "rb") as f:
                self.read_data(f)
        return self._data

    @property
    def analysis(self):
        """Get ANALYSIS segment of the FCS file."""
        if self._analysis is None:
            with open(self.path, "rb") as f:
                self.read_analysis(f)
        return self._analysis

    def reformat_meta(self):
        """Collect the meta data information in a more user friendly format.

        Function looks through the meta data, 
        collecting the channel related information into a
        dataframe and moving it into the _channels_ key.
        """
        meta = self.annotation  # For shorthand (passed by reference)
        channel_properties = []

        for key, value in meta.items():
            if key[:3] == "$P1":
                if key[3] not in string.digits:
                    channel_properties.append(key[3:])

        # Capture all the channel information in a list of lists --
        # used to create a data frame
        channel_matrix = [
            [meta.get("$P{0}{1}".format(ch, p)) for p in channel_properties]
            for ch in self.channel_numbers
        ]

        # Remove this information from the dictionary
        for ch in self.channel_numbers:
            for p in channel_properties:
                key = "$P{0}{1}".format(ch, p)
                if key in meta:
                    meta.pop(key)

        num_channels = meta["$PAR"]
        column_names = ["$Pn{0}".format(p) for p in channel_properties]

        df = pd.DataFrame(
            channel_matrix, columns=column_names, index=(1 + numpy.arange(num_channels))
        )

        if "$PnE" in column_names:
            df["$PnE"] = df["$PnE"].apply(lambda x: x.split(","))

        if self._channel_naming not in df.columns.names:
            df[self._channel_naming] = self.get_channel_names()

        df.index.name = 'Channel Number'
        meta['_channels_'] = df
        meta['_channel_names_'] = self.get_channel_names()

    @property
    def dataframe(self):
        """Construct Pandas dataframe."""
        data = self.data
        channel_names = self.get_channel_names()
        return pd.DataFrame(data, columns=channel_names)

def parse(
    path,
    meta_data_only=False,
    compensate=False,
    channel_naming="$PnS",
    reformat_meta=False,
    data_set=0,
    dtype="float32",
    encoding="utf-8",
):
    """Parse an fcs file at the location specified by the path.

    Parameters
    ----------
    path: str
        Path of .fcs file
    meta_data_only: bool
        If True, the parse_fcs only returns the meta_data 
        (the TEXT segment of the FCS file)
    compensate: bool, 
        reserved parameter to indicate whether the  FCS data 
        should be compensated, unimplemented.
    channel_naming: '$PnS' | '$PnN'
        Determines which meta data field is used for naming the channels.
        The default should be $PnS (even though it is not guaranteed to be unique)

        $PnN stands for the short name (guaranteed to be unique).
            Will look like 'FL1-H'
        $PnS stands for the actual name (not guaranteed to be unique).
            Will look like 'FSC-H' (Forward scatter)

        The chosen field will be used to population self.channels

        Note: These names are not flipped in the implementation.
        It looks like they were swapped for some reason 
        in the official FCS specification.
    reformat_meta: bool
        If true, the meta data is reformatted with the channel information organized
        into a DataFrame and moved into the '_channels_' key
    data_set: int
        Index of retrieved data set in the fcs file.
        This value specifies the data set being retrieved from an fcs file 
        with multiple data sets.
    dtype: str | None
        If provided, will force convert all data into this dtype.
        This is set by default to auto-convert to float32 
        to deal with cases in which the original
        data has been stored using a smaller data type (e.g., unit8). 
        This modifies the original
        data, but should make follow up analysis safer in basically all cases.
    encoding: str
        Provide encoding type of the text section.


    Returns
    -------
    if meta_data_only is True:
        meta_data: dict
            Contains a dictionary with the meta data information
    Otherwise:
        a 2-tuple with
            the first element the meta_data (dictionary)
            the second element the data (in either DataFrame or numpy format)

    Examples
    --------
    fname = '../tests/data/EY_2013-05-03_EID_214_PID_1120_Piperacillin_Well_B7.001.fcs'
    meta = parse_fcs(fname, meta_data_only=True)
    meta, data_pandas = parse_fcs(fname, meta_data_only=False)
    """
    if compensate:
        raise ParserFeatureNotImplementedError(
            "Compensation has not been implemented yet."
        )

    read_data = not meta_data_only

    fcs_parser = FCSParser(
        path,
        read_data=read_data,
        channel_naming=channel_naming,
        data_set=data_set,
        encoding=encoding,
    )

    if reformat_meta:
        fcs_parser.reformat_meta()

    meta = fcs_parser.annotation

    if meta_data_only:
        return meta
    else:  # Then include both meta and dataframe.
        df = fcs_parser.dataframe
        df = df.astype(dtype) if dtype else df
        return meta, df